from __future__ import annotations
import json
from .base_connector import BaseConnector

class LocalConnector(BaseConnector):
    def __init__(self, log_file: str):
        """
        Initialize the Local Connector.
        """
        self.log_file = log_file 


    def flush(self, logs: list) -> None:
        """
        Flushes the provided logs into the local log file without loading the file into memory.

        This method appends the new logs directly to the log file in a JSON Lines (jsonl) format. 
        If the log file does not exist, it creates a new log file. This approach allows for efficient 
        appends and easy parsing of individual log entries without needing to load the entire file into memory.

        Args:
            logs (list): A list of log entries to be written to the log file.
        """
        try:
            # Open the file in append mode, creating it if it doesn't exist
            with open(self.log_file, 'a') as f:
                for log in logs:
                    # Write each log entry as a new line in JSON Lines format
                    f.write(json.dumps(log) + '\n')
        except Exception as e:
            print(f'Error appending to file: {e}')



import contextlib
import inspect
import functools
from datetime import datetime
from queue import Queue
import threading
import uuid
import openai
import atexit

from .connectors.local_connector import LocalConnector
from .singleton import Singleton


thread_local_storage = threading.local()

class ExtensiTrace(metaclass=Singleton):
    def __init__(self, client=None, log_file='./event_log.json', connector=None, task_flush_limit=1):
        self.client = client or openai
        self.log_file = log_file
        self.lock = threading.Lock()
        self.agent_id = str(uuid.uuid4())
        self.data_store = dict() 
        self.connector = connector or LocalConnector(log_file)
        self.task_flush_limit = task_flush_limit
        self.task_flush_ids = Queue() 
        self.task_count = 0
        self.to_flush = []
        atexit.register(self.__on_exit)


    def __on_exit(self):
        """
        Method to ensure all remaining logs are flushed upon program interruption or shutdown.
        """
        with self.lock:
            while not self.task_flush_ids.empty():
                task_id = self.task_flush_ids.get()
                self.__add_to_flush(task_id)
            self.__flush_queue()
            self.task_count = 0
            self.data_store = dict()  # Optionally reset the data store if needed
            print("Program interrupted. All pending logs have been flushed.")


    def log(self, track=False):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                if track and hasattr(thread_local_storage, 'task_id') and len(self.data_store[thread_local_storage.task_id]['call_stack']) > 0:
                    raise ValueError("Cannot track a top level function that is already part of a task.")
                if not hasattr(thread_local_storage, 'task_id') or track:
                    thread_local_storage.task_id = str(uuid.uuid4())
                    with self.lock:
                        self.task_flush_ids.put(thread_local_storage.task_id)
                        self.task_count += 1
                        self.data_store[thread_local_storage.task_id] = {
                            'queue': Queue(),
                            'client': self.client,
                            'call_stack': [],
                            'patched': False,
                            'completion_ids': set(),
                            'last_openai_call': {},
                            'metadata': None
                        }

                with self.lock:
                    self.data_store[thread_local_storage.task_id]['call_stack'].append((func.__name__, str(uuid.uuid4())))


                func_args = inspect.signature(func).bind(*args, **kwargs).arguments
                func_args_dict = self.__serialize_arguments(func_args)
                start_time = datetime.now().timestamp()

                with self.__patched_create_method():
                    result = func(*args, **kwargs)

                end_time = datetime.now().timestamp()
                with self.lock:
                    log = self.data_store[thread_local_storage.task_id]['call_stack'].pop()

                    self.__log_event(
                        log_id=log[1],
                        function_name=func.__name__,
                        start_time=start_time,
                        end_time=end_time,
                        args=func_args_dict,
                        result={'result_string':result} if isinstance(result, str) else result,
                        task_id=thread_local_storage.task_id,
                        agent_id=self.agent_id,
                        parent_log_id=self.data_store[thread_local_storage.task_id]['call_stack'][-1][1] if self.data_store[thread_local_storage.task_id]['call_stack'] else None,
                        metadata=self.data_store[thread_local_storage.task_id]['metadata'],
                        inferred_accuracy=None,
                        accuracy_reasoning=None
                    )

                return result
            return wrapper
        return decorator

    
    def add_metadata(self, metadata: dict):
        with self.lock:
            if self.data_store[thread_local_storage.task_id]['metadata']:
                metadata = {**metadata, **self.data_store[thread_local_storage.task_id]['metadata']}
            else:
                self.data_store[thread_local_storage.task_id]['metadata'] = metadata
        return metadata


    # TODO: Known bug is mock create getting called multiple times and throws extra logs
    def mock_create(self, original_create, *args, **kwargs):
        """
        A mock method to wrap around the original create method, logging additional information.
        """
        with self.lock:
            patched = self.data_store[thread_local_storage.task_id]['patched']

        if not patched: 
            chat_args_dict = self.__serialize_arguments(kwargs)
            chat_call_start_time = datetime.now().timestamp()
            result = original_create(*args, **kwargs)  
            chat_call_end_time = datetime.now().timestamp()

            with self.lock:
                self.__log_event(
                    log_id=str(uuid.uuid4()),
                    function_name='openai.chat.completions.create',
                    start_time=chat_call_start_time,
                    end_time=chat_call_end_time,
                    args=chat_args_dict,
                    result=result.model_dump(),
                    task_id=getattr(thread_local_storage, 'task_id', str(uuid.uuid4())),
                    agent_id=self.agent_id,
                    parent_log_id=self.data_store[thread_local_storage.task_id]['call_stack'][-1][1] if self.data_store[thread_local_storage.task_id]['call_stack'] else None,
                    metadata=self.data_store[thread_local_storage.task_id]['metadata'],
                    inferred_accuracy=None,
                    accuracy_reasoning=None
                )
                self.data_store[thread_local_storage.task_id]['patched'] = True

            return result
        else:
            # Directly call the original method if already patched
            return original_create(*args, **kwargs)


    @contextlib.contextmanager
    def __patched_create_method(self):
        """
        A context manager to temporarily patch the create method for logging purposes.
        """
        with self.lock:
            patched = self.data_store[thread_local_storage.task_id]['patched']
        
        if not patched: 
            task_id = thread_local_storage.task_id
            with self.lock:
                original_create = self.data_store[task_id]['client'].chat.completions.create

            def patched_create(*args, **kwargs):
                try:
                    return self.mock_create(original_create, *args, **kwargs)
                finally:
                    # Immediately restore the original method to avoid recursion
                    with self.lock:
                        self.data_store[task_id]['client'].chat.completions.create = original_create
                        self.data_store[thread_local_storage.task_id]['patched'] = False

            with self.lock:
                self.data_store[task_id]['client'].chat.completions.create = patched_create
            try:
                yield
            finally:
                # Ensure the original method is restored if not already done
                with self.lock:
                    if self.data_store[task_id]['client'].chat.completions.create != original_create:
                        self.data_store[task_id]['client'].chat.completions.create = original_create
                        self.data_store[thread_local_storage.task_id]['patched'] = False
        else:
            yield  # If already patched, yield without re-patching


    def __serialize_arguments(self, arguments):
        return {k: self.__serialize_value(v) for k, v in arguments.items()}


    def __serialize_value(self, value):
        if isinstance(value, (str, int, float, bool, type(None))):
            return value
        elif isinstance(value, (list, tuple)):
            return [self.__serialize_value(v) for v in value]
        elif isinstance(value, dict):
            return {k: self.__serialize_value(v) for k, v in value.items()}
        else:
            return str(value)


    def __log_event(self, **log_entry):
        # Example modification to include full call stack in the log
        task_id = thread_local_storage.task_id
        if log_entry['function_name'] == 'openai.chat.completions.create':
            openai_id = log_entry['result']['id']
            self.data_store[task_id]['last_openai_call'][openai_id] = log_entry['log_id'] 

        self.data_store[task_id]['queue'].put(log_entry)

        if len(self.data_store[task_id]['call_stack']) == 0 and self.task_count >= self.task_flush_limit:
            # TODO: to be made non-blocking
            removed = 0
            while removed < self.task_flush_limit: 
                task_id = self.task_flush_ids.get()
                removed += 1
                self.__add_to_flush(task_id)
            self.task_count -= self.task_flush_limit
            self.__flush_queue()


    # TODO: add batch flushing with storing in memory
    def __add_to_flush(self, task_id):
        while not self.data_store[task_id]['queue'].empty():
            log_entry = self.data_store[task_id]['queue'].get()
            # Skip the loop if the log id is not the last openai call
            if log_entry['function_name'] == 'openai.chat.completions.create' and self.data_store[task_id]['last_openai_call'][log_entry['result']['id']] != log_entry['log_id']:
                continue
            self.to_flush.append(log_entry)

    
    def __flush_queue(self):
        if self.to_flush:
            self.connector.flush(self.to_flush)
            self.to_flush.clear()



import threading

class Singleton(type):
    """
    This is a thread-safe implementation of a Singleton metaclass.
    It ensures that a class has only one instance and provides a global point of access to it.
    """

    _instances = {}
    _lock: threading.Lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:
                    instance = super().__call__(*args, **kwargs)
                    cls._instances[cls] = instance
        return cls._instances[cls]


from abc import ABC, abstractmethod
from typing import List


class BaseConnector(ABC):
    def __init__(self):
        """
        Initialize the Connector.
        """
        pass
    
    @abstractmethod
    def flush(self, logs: List):
        """
        This method should be overridden by subclasses.
        """
        raise NotImplementedError("This method must be implemented by the subclass.")



from pymongo import MongoClient
from pymongo.errors import ConnectionFailure, OperationFailure
import atexit

from .base_connector import BaseConnector

class MongoConnector(BaseConnector):
    def __init__(self, connection_string: str, db_name: str, collection_name: str, client: MongoClient=None):
        """
        Initialize the MongoDB Connector using a connection string.
        :param connection_string: MongoDB connection URI.
        :param db_name: Name of the database.
        :param collection_name: Name of the collection.
        """
        self.db_name = db_name
        self.collection_name = collection_name
        self.client = client or MongoClient(connection_string, tlsAllowInvalidCertificates=True)
        try:
            # Verify server connectivity
            self.client.admin.command('ping')
        except ConnectionFailure as e:
            print("Failed to connect to MongoDB:", e)
        except OperationFailure as e:
            print("Authentication failed:", e)
        atexit.register(self.close_connection)

    def close_connection(self):
        """
        Close the MongoDB connection.
        """
        if self.client:
            self.client.close()
            print("MongoDB connection closed.")

    def flush(self, json_data):
        """
        Insert a list of JSON documents into the specified MongoDB collection.
        :param json_data: A JSON document or dict to be inserted.
        """
        try:
            db = self.client[self.db_name]
            collection = db[self.collection_name]
            collection.insert_many(json_data)
            print("Data inserted successfully.")
            return True
        except Exception as e:
            print("An error occurred while inserting data:", e)



import json
import psycopg2
import atexit
from psycopg2 import OperationalError
from extensitrace.model import Task

from .base_connector import BaseConnector

class PostgresConnector(BaseConnector):
    def __init__(self, connection_string: str, table_name: str):
        """
        Initialize the PostgreSQL Connector using a connection string.
        :param connection_string: PostgreSQL connection URI.
        :param table_name: Name of the PostgreSQL table.
        """
        self.connection_string = connection_string
        self.table_name = table_name
        self.connection = None
        try:
            self.connection = psycopg2.connect(self.connection_string)
            self.connection.autocommit = True
            print('Connection successful')
        except OperationalError as e:
            print("Failed to connect to PostgreSQL:", e)
        atexit.register(self.close)

        
    def close(self):
        """
        Close the database connection.
        """
        if self.connection:
            self.connection.close()
            print("Database connection closed.")


    def flush(self, json_data: list) -> bool:
        if not self.connection:
            print("Database connection is not established.")
            return False

        if not json_data:
            print("No data to insert.")
            return False

        tasks = [Task(**item) for item in json_data]

        conn = self.connection
        try:
            cur = conn.cursor()
            insert_query = """
            INSERT INTO logs (log_id, function_name, start_time, end_time, args, result, task_id, agent_id, parent_log_id, metadata, inferred_accuracy, accuracy_reasoning)
            VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
            """
            cur.executemany(insert_query, [(task.log_id, task.function_name, task.start_time, task.end_time, 
                                            json.dumps(task.args), json.dumps(task.result), task.task_id, 
                                            task.agent_id, task.parent_log_id, json.dumps(task.metadata), 
                                            task.inferred_accuracy, task.accuracy_reasoning) for task in tasks])
            conn.commit()
            cur.close()
            return True
        except Exception as e:
            print(f"Error inserting data: {e}")
            return False



from .base_connector import BaseConnector
import requests
import json


class ExtensibleConnector(BaseConnector):
    def __init__(self, endpoint: str="http://127.0.0.1:5000/api/push_tasks"):
        """
        Initialize the Extensible Connector.
        """
        self.credentials = {}
        self.endpoint = endpoint 


    def flush(self, logs: list):
        """
        Flushes the provided logs using the connector.

        Args:
            logs (list): A list of log entries to be flushed.
        """
        try:
            headers = {"Content-Type": "application/json"}
            data = json.dumps(logs)

            response = requests.post(self.endpoint, headers=headers, data=data)
            if response.status_code != 201:
                print(f"Failed to push logs: {response.text}")

        except Exception as e:
            print(f"An error occurred while pushing logs: {e}")
            



from dataclasses import dataclass


@dataclass
class Task:
    log_id: str
    function_name: str
    start_time: float 
    end_time: float 
    args: dict 
    result: dict 
    task_id: str
    agent_id: str
    parent_log_id: str
    metadata: dict 
    inferred_accuracy: float 
    accuracy_reasoning: str


from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='extensitrace',
    version='0.0.1',
    packages=find_packages(include=['extensitrace', 'extensitrace.*']),
    install_requires=required,
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache License 2.0',
    author='Parth Sareen, Omkaar Kamath',
    author_email='parth@extensible.dev, omkaar@extensible.dev',
    description='A logger for tracking your agent workflow',
    url='https://github.com/Extensible-AI/extensilog/',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
)
